/*
	windowsreen V6. testprogram for BWindowScreen!!
	by Rudolf Cornelissen 4/2003

	verified for working correctly with the (open)BeOS MAtrox MGA driver V0.14 using R5.0.3PE:
	tested on Mil I upto and including G550 cards.
	also verified for working correctly on the Be Diamand Speedstar A70 (Sis6326DVD) driver:
	including TWEAK_METHOD: both setups are 100% OK.
*/
/*
	Note please:
	Programming C++ is not something I am very good in ATM, so if you encounter me not
	destroying or closing down objects on quitting for example, please take the nessesary
	steps yourself. Thanks.
	
	Rudolf.

*/
#include <stdio.h>
#include <be/game/WindowScreen.h>
#include <be/app/Application.h>
#include <be/interface/Screen.h>
#include <Bitmap.h>

/* Matrox driver dualhead extensions to flags */
#define DUALHEAD_OFF (0<<6)
#define DUALHEAD_CLONE (1<<6)
#define DUALHEAD_ON (2<<6)
#define DUALHEAD_SWITCH (3<<6)
#define DUALHEAD_BITS (3<<6)
#define DUALHEAD_CAPABLE (1<<8)
#define TV_BITS (3<<9)
#define TV_MON (0<<9)
#define TV_PAL (1<<9)
#define TV_NTSC (2<<9)
#define TV_CAPABLE (1<<11)

/*
	you can uncomment the next line if you want to see how I illegally set virtual_width to 1280
	if BWindowScreen.SetFrameBuffer cannot be used..
	Works perfectly on the (open)BeOS matrox driver!!
*/
//#define TWEAK_METHOD

int main()
{
	bool dual = false, tv = false;
	unsigned char modeNr;
	int getmode = true;
	const char *name;
	uint16 virt_width = 1280;
	graphics_card_info *info;

	status_t status = B_ERROR;
	name = "testje";

	new BApplication("application/x.vnd.rudscreen");
	printf("BWindowScreen testapp V6\n\n");

	display_mode mode, orgmode, testmode;
	/*
	 	Technically (speaking from the videodriver's point of view) there is absolutely no reason
	 	whatsoever why the BWindowScreen class, which _is_ already derived from BScreen, should
	 	not be able to perform the SetMode and Getmode commands for instance.
	 	
	 	Personally I am having trouble understanding why I can't just do BWindowScreen.SetMode()
	 	to set any BWindowScreen mode I want!
	 	I strongly advice people to modify this behaviour in future versions of BeOS ;-)
		And make sure you keep BWindowScreen! It's an important class!
	 	
	 	Anyway: I want this, and I know the drivers should have no trouble with it.
	 	So:
	 	Let's first get a handle to BScreen, then startup BWindowScreen, and finally use both
	 	intermixed to do exactly what we want...
	 	Just make sure that in the end you reset BScreen to the original mode or BeOS it's
	 	app_server will get a bit confused.
	*/
	BScreen *myRealScreen = new BScreen();
	if (!myRealScreen->IsValid())
	{
		printf("\nCould not get handle to Bscreen, exiting...\n");
		return(B_OK);
	}

	/* 
		save org mode so we will remain friends with the app_server
	 */
	if (myRealScreen->GetMode(&orgmode) != B_OK)
	{
		getmode = false;
	}

	//show menu:
	printf("\nSelect Matrox output mode:\n");
	printf("1. Singlehead (std) mode (1024 x 768 x 16)\n");
	//BTW: in 16bit color Matrox does dualhead modes with stepgranularity = 32 pixels!!
	//(16pixels in 32bit mode..)
	printf("2. Dualhead clone mode (1024 x 768 x 16)\n");
	printf("3. Dualhead stretch mode (1024 x 768 x 16)\n");
	printf("4. Dualhead switch mode (1024 x 768 x 16)\n");
	printf("5. NTSC TVout clone mode (1024 x 768 x 16)\n");
	printf("6. PAL TVout clone mode (1024 x 768 x 16)\n");
	printf("7. NTSC TVout stretch mode (1024 x 768 x 16)\n");
	printf("8. PAL TVout stretch mode (1024 x 768 x 16)\n");
	printf("9. NTSC TVout switch mode (1024 x 768 x 16)\n");
	printf("0. PAL TVout switch mode (1024 x 768 x 16)\n");
	printf("Option: ");
	fflush(stdin);
	do scanf("%c",&modeNr);
	while ((modeNr < '0') | (modeNr > '9'));
	modeNr -= 48;

	/*
		now set the BWindowScreen mode that's closest to the mode you want
	 */
	BWindowScreen *myScreen ;
	myScreen = new BWindowScreen(name, B_16_BIT_1024x768, &status, false);
	if (status == B_OK)
		printf("constuction status: B_OK\n");
	else
	{
		printf("constuction status: B_ERROR\n");
    	myScreen->PostMessage(B_QUIT_REQUESTED, myScreen);
	}

	if (myScreen->CanControlFrameBuffer())
		printf("CanControlFrameBuffer: Yes\n");
	else
	{
		printf("CanControlFrameBuffer: No\n");
		return B_OK;
	}

	/* let's see what we can do */
	if (myRealScreen->GetMode(&testmode) == B_OK)
	{
		/* check if we have a dualhead and/or TVout capable card (no range checking) */
		myRealScreen->ProposeMode(&testmode, &testmode, &testmode);
		if (testmode.flags & DUALHEAD_CAPABLE)
		{
			dual = true;
			printf("Dualhead capable card with ditto mode detected\n");
		}
		else printf("Suppressing dualhead options.\n");
		if (testmode.flags & TV_CAPABLE)
		{
			tv = true;
			printf("TVout capable card with ditto mode detected\n");
		}
		else printf("Suppressing TVout options.\n");
	}
	else
	{
		printf("Getmode failed\n");
	}

	/* if we use stretch or switch dualhead modes, the workspace is twice the size:
	 Note that we also take 512 pixels offscreen instead of 256 in this example. */
	if (dual)
		switch (modeNr)
		{
		case 3:
		case 4:
		case 7:
		case 8:
		case 9:
		case 0:
			virt_width = 1280 * 2;
			break;
		}	

	/* 
		if you want for instance a VCD or DVD mode that also works on TV, you can now issue
		a BScreen.Setmode() command to do it.
		This is unofficial BTW, but on the (open)BeOS Matrox driver at least this is working
		perfectly.
		This means you can of course also set any refreshrate you want! 
		You are _not_ restricted to 60Hz only!
		(tested on:
		 Be Matrox Millenium driver, and on Be Diamand speedstar A70 (Sis6326DVD) driver)
	*/
	//set up a mode here, then:
	//myRealScreen->Setmode(&mode, false);

	/* displaying rubbish now: */
	myScreen->Show();

	/*
		Now, you have two choices:
	 	-	Officially (recommended), use BWindowScreen.Setframebuffer() to set the virtual
	 	 	screen size;
			Unfortunately, there's a bug in BWindowScreen that might let this function fail
			if the driver does not take precautions. The (open)BeOS Matrox driver takes these
			precautions BTW.

	 	-	If the official method does not work, or you are using a custom display_mode
	 	    activated via BScreen after setting the initial mode with BWindowScreen,
	 	    you can use BScreen.Setmode() to set the virtual screen size.
			Because this is not official, send it a temporary COPY of the mode to set as
			BScreen tampers with it! (virtual_width tampering confirmed!)
	*/
	/*
		wait until we are absolutely sure myScreen is connected 
 		BEFORE calling SetFrameBuffer!
 		otherwise the program may hang...
 		(another bug in BeOS??)
 		
 		BTW: I heard you can just poll for the BWindowScreen status before you continue...
 	 */
	snooze(100000);
	/* using SetFrameBuffer because we use a standard BWindowScreen mode currently */
#ifdef TWEAK_METHOD
{
}
#else
{
	if ((myScreen->SetFrameBuffer(virt_width, 768)) != B_OK)
	{
		printf("SetFrameBuffer: error..\n");
    	myScreen->PostMessage(B_QUIT_REQUESTED, myScreen);

		if (getmode)
			myRealScreen->SetMode(&orgmode, false);
		printf("BWindowScreen trouble:\nYou could try to recompile with TWEAK_METHOD enabled,"
			"and run this app again.\nExiting...\n");
		return(B_OK);
	}
	else
		printf("SetFrameBuffer to virtual_width = %d: done.\n", virt_width);
}
#endif
	/* BTW: OK, we are still 60Hz: check your monitor timing to verify. I'll wait a bit...*/
	snooze(10000000);

	/*
		Hey! Let's use BScreen now.
		let's do 70 Hz, (and while we are at it, you could change the picture width)
	 */
	if (myRealScreen->GetMode(&mode) == B_OK)
	{
		/* refresh = 70 Hz */
		mode.timing.pixel_clock = (mode.timing.h_total * mode.timing.v_total * 70)/1000;

		/* 
			OK, you could use another way (so not via SetFrameBuffer) of setting the virtual
			picture	size we want:
			Beware:
			use a _copied_ mode!! 
			BScreen resets virtual_width to 1024, I suppose app_server is a bit confused...
		 	setmode does setup 1280 though.. 
		 */
#ifdef TWEAK_METHOD
{
		mode.virtual_width = virt_width;
		printf("Image width set to = %d, height = %d\n",mode.virtual_width,mode.virtual_height);
}
#endif
		/* the driver will automatically suppress TVout or dualhead if card not capable! */
		switch (modeNr)
		{
		case 1:
			/* std mode: nothing to do here */
			break;
		case 2:
			mode.flags &= ~(DUALHEAD_BITS | TV_BITS);
			mode.flags |= (TV_MON | DUALHEAD_CLONE);
			break;
		case 3:
			mode.flags &= ~(DUALHEAD_BITS | TV_BITS);
			mode.flags |= (TV_MON | DUALHEAD_ON);
			break;
		case 4:
			mode.flags &= ~(DUALHEAD_BITS | TV_BITS);
			mode.flags |= (TV_MON | DUALHEAD_SWITCH);
			break;
		case 5:
			mode.flags &= ~(DUALHEAD_BITS | TV_BITS);
			mode.flags |= (TV_NTSC | DUALHEAD_CLONE);
			break;
		case 6:
			mode.flags &= ~(DUALHEAD_BITS | TV_BITS);
			mode.flags |= (TV_PAL | DUALHEAD_CLONE);
			break;
		case 7:
			mode.flags &= ~(DUALHEAD_BITS | TV_BITS);
			mode.flags |= (TV_NTSC | DUALHEAD_ON);
			break;
		case 8:
			mode.flags &= ~(DUALHEAD_BITS | TV_BITS);
			mode.flags |= (TV_PAL | DUALHEAD_ON);
			break;
		case 9:
			mode.flags &= ~(DUALHEAD_BITS | TV_BITS);
			mode.flags |= (TV_NTSC | DUALHEAD_SWITCH);
			break;
		case 0:
			mode.flags &= ~(DUALHEAD_BITS | TV_BITS);
			mode.flags |= (TV_PAL | DUALHEAD_SWITCH);
			break;
		}

		printf("Now calling BScreen.Setmode!\n");
		myRealScreen->SetMode(&mode, false);
		printf("refresh = 70Hz\n");
		/* BTW: OK, we are now in the dualhead mode if requested. I'll wait a bit...*/
		snooze(10000000);

#ifdef TWEAK_METHOD
{
		printf("Hey!! BScreen says it reset my mode to the old virtual width (it's tampering!!)\n");
		printf("Image width = %d, height = %d\n",mode.virtual_width,mode.virtual_height);

		/* just verifying we are not stupid about that virtual_width tampering 'bug'.. */	
		myRealScreen->GetMode(&mode);
		printf("Image width after verifying the just set mode: via a BScreen.GetMode() call:\n"
			"virtual width = %d, height = %d\n",mode.virtual_width,mode.virtual_height);
		printf("But:\nWe _know_ we have a virtual_width of %d. Right? Right!\n", virt_width);
}
#endif
	}
	else
	{
		printf("Getmode failed\n");
	}

	/* We should (also) be on 70Hz now.. See the monitors onscreen settings */

	/*
		Now: let's draw something just to be sure we see something (vertical bar):
		BUT: make sure you don't go outside the buffer you got!! 
		(or your app will crash at the least)
	 */
	info = myScreen->CardInfo();
	/* clear screen: doing 4bytes per write because of speed... */
	for (uint32 *a = (uint32 *)(*info).frame_buffer; a < (uint32 *)(*info).frame_buffer +
														 (655362 * ((uint32)(virt_width / 1280))) - 1; a++)
	{
		*a = 0x00000000;
	}
	/* now draw a vertical yellow line */
	for (uint8 *a = (uint8 *)(*info).frame_buffer + 999; a < (uint8 *)(*info).frame_buffer +
												 (1024000 * ((uint32)(virt_width / 1280))); a+= virt_width * 2)
	{
		*a = 0xff;
		*(a + 2) = 0xff;
	}

	/* now draw a second vertical line (blue) if dualhead stretch or switch mode active */
	if (dual)
		switch (modeNr)
		{
		case 3:
		case 4:
		case 7:
		case 8:
		case 9:
		case 0:
			for (uint8 *a = (uint8 *)(*info).frame_buffer + 3000; a < (uint8 *)(*info).frame_buffer +
												 (1024000 * ((uint32)(virt_width / 1280))); a+= virt_width * 2)
			{
				*a = 0xff;
				*(a + 2) = 0xff;
			}
			break;
		}

	/* let's wait a while */
	snooze(5000000);

	/*
		start rolling the screen.. The driver will note every pixelstep, but it will only actual
		step once the stepgranularity for the mode set is reached: 
			8bit color: 8 pixels
			16bit color: 4 pixels 
			32bit color: 2 pixels 
 		We are in 16bit color, so every four loops below you will see the display move 4 pixels at once. 
		this applies for MIL1 upto and including G550.
		(BTW: Nvidia TNT1 and up (for instance) can always do single pixelsteps here!)

		Matrox secondary head granularity:
			16bit color: 32 pixels
			32bit color: 16 pixels
	 */

	/*
		BTW: You cannot know what granularity your driver/card has. If you want to know you are
		doing an exaxt setup, make sure you keep the lowest bits in the step clear.
		Matrox cards do worst case steps of 32, so you would need to keep b0-4 zero at least!
		
		Doing an exact setup will be important if you are doing video output with buffer_flipping
		with BWindowScreen for instance!


		Note: 
		Be planned to setup a new accelerant hook to overcome this problem. AFAIK it was never
		implemented: (should be confirmed though I would say)
		
		taken from:
			BeOS R4 Graphics Driver Kit.
			(alpha release 2, 1999-03-30)
		written by:
		Trey Boudreau (AFAIK)
	
	#if NEW_FOR_R4.1
		B_GET_MOVE_DISPLAY_CONSTRAINTS  -  No feature specific data required.
		typedef status_t (*get_move_display_constraints)
			(display_mode *dm, uint32 *pixels, uint32 *lines);

		For the specified display_mode, return the smallest number of pixels the display
		can be moved in the horizontal direction and the smallest number of lines the display
		can be moved in the vertical direction.  This is essentially the granularity of the
		move_display_area() feature.
	#endif
 	 */
	for (int cnt = 0; cnt <= 256; cnt += 1)
	{
		if (myScreen->MoveDisplayArea(cnt, 0) != B_OK)
		{
			printf("MoveDisplay error!\n");
			/* for exiting the loop */
			cnt = 260;
		}
		snooze (10000);
	}

	/* and back.. */
	for (int cnt = 256; cnt >= 0; cnt -= 1)
	{
		if (myScreen->MoveDisplayArea(cnt, 0) != B_OK)
		{
			printf("MoveDisplay error!\n");
			/* for exiting the loop */
			cnt = -1;
		}
		snooze (10000);
	}

	/* disable BWindowScreen.  */
	myScreen->Lock();
	myScreen->Quit();

	/* reset screen to original Desktop mode */
	if (getmode)
		myRealScreen->SetMode(&orgmode, false);

	delete(myRealScreen);
	/* hmm, apparantly myScreen was already deleted. If I enable the next line, BeOS
	 * might hangs this app (or the terminal session?) on exiting below. */
	//delete(myScreen);
	
	printf("\nDone! Exiting...\n");
	return B_OK;
}
